/*-
 * Copyright 1999, 2001 John-Mark Gurney.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: main.c,v 1.16 2001/03/05 01:01:40 jmg Exp $
 *
 */

#ifdef __sun__
#undef __STDC__
#define __STDC__ 0
#endif
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#ifdef __FreeBSD__
#include <ufs/ffs/fs.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#include <ufs/ufs/dir.h>
#define DI_SPARE
#define DI_INUMBER
#define DI_NSEC
#define FS_INODEFMT
#define REAL_QUAD_T
#elif defined(__sun__)
#include <sys/fs/ufs_fs.h>
#include <sys/fs/ufs_quota.h>
#include <sys/fs/ufs_inode.h>
#include <sys/fs/ufs_fsdir.h>
typedef daddr32_t ufs_daddr_t;
#define di_flags	di_ic.ic_flags
#define	di_rdev		di_ordev
/* solaris open doesn't support locking */
#define O_EXLOCK	0
#endif
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#ifdef __FreeBSD__
#define PF_LLONG	"q"
#elif defined(__sun__)
#define PF_LLONG	"ll"
#endif

/* lets use a default block size of 64k for basic operations */
/* must be a multiple of DEV_BSIZE */
/* right now ata can transfer 128k at once, make use of this */
#define INT_BLKSIZE	(256 * DEV_BSIZE)

struct proginfo {
	struct	fs *fs;
	int	fd;
	int	seekable;
	void	*map;
	struct	stat st;
	off_t	off;
};

int readin(struct proginfo pi, void *data, off_t pos, size_t len);
void *areadin(struct proginfo pi, off_t pos, size_t len);
struct fs *readinsb(struct proginfo pi, off_t blk);
off_t getblkoff(struct proginfo pi, ufs_daddr_t blk);
off_t inumtooff(struct proginfo pi, ino_t inum);
void writeout(struct proginfo pi, int fd, off_t pos, size_t len);
ufs_daddr_t getblknuminode(struct proginfo pi, struct dinode inode,
	uint64_t blk);
void checksbcg(struct proginfo pi, off_t pos);

/* more generic functions */
int bpread(int fd, void *data, size_t len, off_t off);
void printsuper(struct fs *fs);
char *validatesb(struct fs *fs);
void printdinode(struct dinode inode);

void findsuperblks(struct proginfo pi);
void getinode(struct proginfo pi, void *data, struct dinode ino);
void printdirect(struct direct *dir, int size);
ino_t *matchname(struct proginfo pi, struct dinode dir, char *name);
void searchfsforname(struct proginfo pi, char *name);
void printdirs(struct proginfo pi);
void dumpoutinode(struct proginfo pi, ino_t num);
void dumpall(struct proginfo pi);
void printinodeinfo(struct proginfo pi, ino_t inum);
const char *typestr(int type);
void duplicateinode(struct proginfo pi, ino_t inum, const char *name);
int inodetofile(struct proginfo pi, struct dinode inode, const char *name);
void inodetofd(struct proginfo pi, struct dinode inode, int fd);
int isblkininode(struct proginfo pi, ufs_daddr_t blk, struct dinode inode);

int pow(int a, int b);

#define FBLK_ERR	-1
#define FBLK_CG		1
#define FBLK_INODE	2
void findblk(struct proginfo pi, ufs_daddr_t blk, ino_t *num);

void
usage()
{
	printf("usage: ffsrecover [-adps] [-b blk] [-c inum -n name | -f inum | -i inum | -b blknum] [-o blk] [-r name] [-t blk] file\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;
	struct proginfo pi;
	char *file;
	char *t;
	int ch, findsb, printbsb, recoverfile, printalldirs;
	ufs_daddr_t blk;
	ino_t inum;
	int dumpinode;
	int dumpallfiles;
	int printinfo;
	int createinode;
	int fblk;
	off_t sblk;
	off_t startblk;

	findsb = 0;
	printbsb = 0;
	recoverfile = 0;
	printalldirs = 0;
	dumpinode = 0;
	dumpallfiles = 0;
	printinfo = 0;
	createinode = 0;
	fblk = 0;
	inum = 0;
	file = NULL;
	blk = 0;
	sblk = SBLOCK;
	startblk = 0;

	while ((ch = getopt(argc, argv, "ab:c:df:i:n:o:pr:st:")) != -1)
		switch (ch) {
		case 'a':
			dumpallfiles = 1;
			break;
		case 'b':
			fblk = 1;
			blk = atoi(optarg);
			break;
		case 'c':
			createinode = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'd':
			printalldirs = 1;
			break;
		case 'f':
			printinfo = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'i':
			dumpinode = 1;
			if (inum)
				usage();
			inum = atoi(optarg);
			break;
		case 'n':
			if (file != NULL)
				usage();
			file = strdup(optarg);
			break;
		case 'o':
			sblk = strtoq(optarg, NULL, 0);
			break;
		case 'p':
			printbsb = 1;
			break;
		case 'r':
			recoverfile = 1;
			file = strdup(optarg);
			break;
		case 's':
			findsb = 1;
			break;
		case 't':
			startblk = strtoq(optarg, NULL, 0);
			break;
		case '?':
		default:
			usage();
		}

	argc -= optind;
	argv += optind;

	if (argc == 0 || argc > 1)
		usage();

	if (createinode && file == NULL)
		usage();

	if ((pi.fd = open(*argv, O_RDONLY, 0)) == -1) {
		perror("open");
		exit(1);
	}

	if (lseek(pi.fd, 0, SEEK_CUR) == -1 && errno == ESPIPE)
		pi.seekable = 0;
	else
		pi.seekable = 1;

	if (fstat(pi.fd, &pi.st) == -1) {
		perror("stat");
		exit(1);
	}

	pi.off = startblk * DEV_BSIZE;
	if ((pi.map = mmap(0, pi.st.st_size - pi.off, PROT_READ, MAP_PRIVATE,
	    pi.fd, pi.off)) == MAP_FAILED) {
		fprintf(stderr, "mmap failed, reverting to read\n");
		pi.map = NULL;
	}

	pi.fs = readinsb(pi, sblk);
	if ((t = validatesb(pi.fs)) != NULL)
		fprintf(stderr,
		   "WARNING: field %s is invalid, superblock may be corrupt.\n",
		    t);

	if (pi.fs->fs_magic != FS_MAGIC)
		fprintf(stderr, "doesn't have magic\n");

	if (createinode)
		duplicateinode(pi, inum, file);

	if (dumpallfiles)
		dumpall(pi);

	if (dumpinode)
		dumpoutinode(pi, inum);

	if (printalldirs)
		printdirs(pi);

	if (findsb)
		findsuperblks(pi);

	if (printinfo)
		printinodeinfo(pi, inum);

	if (printbsb)
		printsuper(pi.fs);

	if (recoverfile)
		searchfsforname(pi, file);

	if (fblk)
		findblk(pi, blk, &inum);

	return 0;
}

void
findsuperblks(struct proginfo pi)
{
	off_t *sblocks;
	int32_t *sbmagic;
	int sblockscnt;
	off_t i;
	off_t sbpos;
	int moff;
	int mdif;
	int32_t *tm;
	off_t end;
	int sr;

	sblocks = NULL;
	sblockscnt = 0;

	if ((sbmagic = malloc(INT_BLKSIZE)) == NULL) {
		perror("malloc");
		exit(1);
	}

	moff = ((int)(&((struct fs *)NULL)->fs_magic)) % DEV_BSIZE;
	mdif = ((int)(&((struct fs *)NULL)->fs_magic)) & -DEV_BSIZE;

	if (pi.st.st_size == 0 || (pi.st.st_mode & S_IFMT) == S_IFCHR)
		end = (off_t)LLONG_MAX;
	else
		end = pi.st.st_size;

	for (i = 0; i < end; i += DEV_BSIZE) {
		if ((i % INT_BLKSIZE) == 0)
			if ((sr = readin(pi, sbmagic, i, INT_BLKSIZE))
			    != INT_BLKSIZE)
				end = i + sr;
		tm = (int32_t *)((void *)sbmagic + i % INT_BLKSIZE + moff);
		if (*(int32_t *)((void *)sbmagic + i % INT_BLKSIZE + moff)
		    == FS_MAGIC) {
			sbpos = i - mdif;
			if ((sblocks = realloc(sblocks, ++sblockscnt
			    * sizeof *sblocks)) == NULL) {
				perror("realloc");
				exit(1);
			}
			sblocks[sblockscnt - 1] = sbpos;
			checksbcg(pi, sbpos);
			fflush(stdout);
		}
	}

	free(sbmagic);
	free(sblocks);
}

void
checksbcg(struct proginfo pi, off_t pos)
{
	struct fs *fs;
	struct cg *cg;
	off_t cgpos;
	off_t fsb;
	off_t sboff;

	fs = areadin(pi, pos, SBSIZE);
	if (fs->fs_magic != FS_MAGIC)
		goto nocg;

	printf("superblock: %qd(byte), %qd(block)",
	    (int64_t)pos, (int64_t)pos / DEV_BSIZE);

	sboff = fsbtodb(fs, fs->fs_sblkno) * DEV_BSIZE;
	cgpos = pos + fsbtodb(fs, fs->fs_cblkno) * DEV_BSIZE - sboff;
	/*
	 * this is to prevent outragious struct fs's from causing problems,
	 * no cg should be more than 64k, though this can still cause FP
	 * issues.
	 */
	cg = NULL;
	if (CGSIZE(fs) < 64 * 1024) {
		cg = areadin(pi, cgpos, CGSIZE(fs));

		if (cg->cg_magic != CG_MAGIC) {
			sboff = SBOFF;
			cgpos = pos + fsbtodb(fs, fs->fs_cblkno) * DEV_BSIZE
			    - sboff;
			cg = areadin(pi, cgpos, CGSIZE(fs));
		}

		if (cg->cg_magic == CG_MAGIC) {
			printf(", cg: %qd(byte), %qd(block), %d(nth)", cgpos,
			    cgpos / DEV_BSIZE, cg->cg_cgx);
			fsb = (pos - sboff) / DEV_BSIZE - fsbtodb(fs,
			    cgbase(fs, cg->cg_cgx));
			printf(
			    ", fs begin: %qd(block), primary sb at: %qd(block)",
			    fsb, fsb + SBOFF / DEV_BSIZE);
		}
	}

	printf("\n");

	free(cg);
nocg:
	free(fs);
}

void
getinode(struct proginfo pi, void *data, struct dinode inode)
{
	uint64_t i;
	ufs_daddr_t blktoget;

	for (i = 0; i <= (inode.di_size >> pi.fs->fs_bshift); i++) {
		if ((blktoget = getblknuminode(pi, inode, i)) == 0)
			memset(data + lblktosize(pi.fs, i), 0, dblksize(pi.fs,
			    &inode, i));
		else
			readin(pi, data + lblktosize(pi.fs, i), getblkoff(pi,
			    blktoget), dblksize(pi.fs, &inode, i));
	}
}

off_t
getblkoff(struct proginfo pi, ufs_daddr_t blk)
{
	off_t a, b;

	a = fsbtodb(pi.fs, blk);
	b = a * DEV_BSIZE;

	return b;
}

void
printdirect(struct direct *dir, int size)
{
	struct direct *ptr;

	for(ptr = dir; (char *)ptr < (char *)dir + size;
	    ((char *)ptr) += ptr->d_reclen)
		printf("(%d, \"%s\"), ", ptr->d_ino, ptr->d_name);
}

void
printdirs(struct proginfo pi)
{
	struct dinode inode;
	struct direct *dir;
	int i;

	for (i = 2; i < pi.fs->fs_ncg * pi.fs->fs_ipg; i++) {
		if (readin(pi, &inode, inumtooff(pi, i), sizeof inode)
		    != sizeof inode) {
			fprintf(stderr, "short read on inode %d, skipping\n",
			    i);
			continue;
		}

		if (i == 2 || (inode.di_mode & IFMT) == IFDIR) {
			if ((dir = malloc(blkroundup(pi.fs, inode.di_size))) == NULL) {
				fprintf(stderr,
				    "memory allocation failed for inode %d\n",
				    i);
				continue;
			}
			getinode(pi, dir, inode);
			printf("inode %d: ", i);
			printdirect(dir, inode.di_size);
			printf("\n");
			free(dir);
		}
	}
}

void
searchfsforname(struct proginfo pi, char *name)
{
	struct dinode inode;
	struct direct *dir;
	int i, j;
	ino_t *inol;

	for (i = 2; i < pi.fs->fs_ncg * pi.fs->fs_ipg; i++) {
		if (readin(pi, &inode, inumtooff(pi, i), sizeof inode)
		    != sizeof inode) {
			fprintf(stderr, "short read on inode %d, skipping\n",
			    i);
			continue;
		}
		if (i == 2 || (inode.di_mode & IFMT) == IFDIR) {
			if ((dir = malloc((inode.di_size + pi.fs->fs_fsize - 1)
			    & pi.fs->fs_fmask)) == NULL) {
				fprintf(stderr,
				    "memory allocation failed for inode %d\n",
				    i);
				continue;
			}
			getinode(pi, dir, inode);
			inol = matchname(pi, inode, name);
			for (j = 0; inol[j] != 0; j++)
				printf("matched: %6d\tparent: %6d\n", inol[j], i);
			free(inol);
			free(dir);
		}
	}
}

ino_t *
matchname(struct proginfo pi, struct dinode dir, char *name)
{
	ino_t *list;
	ino_t *nlist;
	int lalloc;
	void *data;
	struct direct *ptr;

	if ((data = malloc((dir.di_size + pi.fs->fs_fsize - 1)
	    & pi.fs->fs_fmask)) == NULL) {
		fprintf(stderr, "memory allocation failed for directory\n");
		return NULL;
	}
	if ((list = malloc(sizeof *list * (lalloc = 1))) == NULL) {
		fprintf(stderr, "memory allocation failed for directory\n");
		return NULL;
	}

	getinode(pi, data, dir);

	list[lalloc - 1] = 0;
	for (ptr = (struct direct *)data; (void *)ptr < data + dir.di_size;
	    ((char *)ptr) += ptr->d_reclen)
		if (strcmp(ptr->d_name, name) == 0) {
			if ((nlist = realloc(list, ++lalloc)) == NULL) {
				fprintf(stderr,
				    "can't alocate more memory, aborting\n");
				free(data);
				free(list);
				return NULL;
			}
			list = nlist;
			list[lalloc - 2] = ptr->d_ino;
			list[lalloc - 1] = 0;
		}
	free(data);

	return list;
}

off_t
inumtooff(struct proginfo pi, ino_t inum)
{
	off_t a, b, c, d;

	a = ino_to_fsbo(pi.fs, inum) * sizeof(struct dinode);
	b = fsbtodb(pi.fs, ino_to_fsba(pi.fs, inum));
	c = b * DEV_BSIZE;
	d = c + a;
	return d;
}

/*
 * XXX - for large files this is not optimal, should make inode block
 * traversal a function, and then interate over it here
 */
void
dumpoutinode(struct proginfo pi, ino_t num)
{
	struct dinode inode;
	void *data;

	if (readin(pi, &inode, inumtooff(pi, num), sizeof inode)
	    != sizeof inode) {
		fprintf(stderr, "short read on inode %d, skipping\n",
		    num);
		return;
	}

	if ((data = malloc(blkroundup(pi.fs, inode.di_size))) == NULL) {
		fprintf(stderr, "can't allocate memory for %d\n", num);
		return;
	}
	getinode(pi, data, inode);

	write(STDOUT_FILENO, data, inode.di_size);

	free(data);
}

void
dumpall(struct proginfo pi)
{
	struct dinode inode;
	void *data;
	int i;
	char name[16];

	for (i = 2; i < pi.fs->fs_ncg * pi.fs->fs_ipg; i++) {
		if (readin(pi, &inode, inumtooff(pi, i), sizeof inode)
		    != sizeof inode) {
			fprintf(stderr, "short read on inode %d, skipping\n",
			    i);
			continue;
		}
		if ((inode.di_mode & IFMT) == IFREG) {
			if ((data = malloc((inode.di_size + pi.fs->fs_fsize
			    - 1) & pi.fs->fs_fmask)) == NULL) {
				fprintf(stderr,
				    "memory allocation failed for inode %d\n",
				    i);
				continue;
			}
			getinode(pi, data, inode);
			snprintf(name, sizeof name, "%d", i);
			inodetofile(pi, inode, name);
			free(data);
		}
	}
}

void
printinodeinfo(struct proginfo pi, ino_t inum)
{
	struct dinode inode;

	if (readin(pi, &inode, inumtooff(pi, inum), sizeof inode)
	    != sizeof inode) {
		fprintf(stderr, "short read on inode %d, skipping\n",
		    inum);
		return;
	}

	printf("inum: %d\n", inum);
	printdinode(inode);
}

void
printdinode(struct dinode inode)
{
	printf("mode: %s\n", typestr(inode.di_mode & IFMT));
	printf("perm: %o\n", inode.di_mode);
	printf("link: %d\n", inode.di_nlink);
	printf("size: %qd\n", inode.di_size);
	printf("mtim: %s", ctime((long *)&inode.di_mtime));
	printf("ctim: %s", ctime((long *)&inode.di_ctime));
	printf("atim: %s", ctime((long *)&inode.di_atime));
	printf("flgs: %d\n", inode.di_flags);
	printf("blks: %d\n", inode.di_blocks);
	printf(" gen: 0x%x\n", inode.di_gen);
	printf(" uid: %d\n", inode.di_uid);
	printf(" gid: %d\n", inode.di_gid);
	printf("rdev: 0x%x\n", inode.di_rdev);
#ifdef DI_SPARE
	printf("spr1: 0x%x\n", inode.di_spare[0]);
	printf("spr2: 0x%x\n", inode.di_spare[1]);
#endif
#ifdef DI_INUMBER
	printf("lfsi: 0x%x\n", inode.di_inumber);
#endif
}

const char *
typestr(int type)
{
	static char buf[32];
	if (type == IFIFO)
		return "pipe";
	if (type == IFCHR)
		return "character";
	if (type == IFDIR)
		return "directory";
	if (type == IFBLK)
		return "block";
	if (type == IFREG)
		return "regular";
	if (type == IFLNK)
		return "symlink";
	if (type == IFSOCK)
		return "socket";
	snprintf(buf, sizeof buf, "invalid type 0x%x", type);
	return buf;
}

int
pow(int a, int b)
{
	int res;

	for (res = 0; b > 0; b--)
		res *= a;

	return res;
}

void
duplicateinode(struct proginfo pi, ino_t inum, const char *name)
{
	struct timeval times[2];
	struct direct *dir, *ptr;
	struct dinode inode;
	mode_t oldumask;
	int failed;

	if (readin(pi, &inode, inumtooff(pi, inum), sizeof inode)
	    != sizeof inode) {
		fprintf(stderr, "short read on inode %d, skipping\n",
		    inum);
		return;
	}
	failed = 0;

	oldumask = umask(0);
	switch(inode.di_mode & IFMT) {
	case IFREG:
		inodetofile(pi, inode, name);
		break;
	case IFDIR:
		if (mkdir(name, inode.di_mode & 07777) == -1) {
			failed = 1;
			break;
		}
		if (chdir(name) == -1) {
			failed = 1;
			break;
		}
		if ((dir = malloc(blkroundup(pi.fs, inode.di_size))) == NULL) {
			fprintf(stderr,
			    "memory allocation failed for inode %d\n",
			    inum);
			failed = 1;
			break;
		}
		getinode(pi, dir, inode);
		for(ptr = dir; (char *)ptr < (char *)dir + inode.di_size;
		    ((char *)ptr) += ptr->d_reclen)
			duplicateinode(pi, ptr->d_ino, ptr->d_name);
		free(dir);
		if (chdir("..") == -1) {
			failed = 1;
			break;
		}
		break;
	case IFSOCK:
		fprintf(stderr, "don't know how to create sockets (%s)\n",
		    name);
		break;
	case IFIFO:
		if (mkfifo(name, inode.di_mode & 07777) == -1)
			failed = 1;
		break;
	case IFLNK:
		fprintf(stderr, "don't know how to create symlinks (%s)\n",
		    name);
		break;
	case IFBLK:
	case IFCHR:
		if (mknod(name, inode.di_mode, inode.di_rdev) == -1)
			failed = 1;
		break;
	default:
		failed = 1;
		break;
	}
	if (!failed) {
		times[0].tv_sec = inode.di_atime;
#ifdef DI_NSEC
		times[0].tv_usec = inode.di_atimensec / 1000;
#else
		times[0].tv_usec = 0;
#endif
		times[1].tv_sec = inode.di_mtime;
#ifdef DI_NSEC
		times[1].tv_usec = inode.di_mtimensec / 1000;
#else
		times[1].tv_usec = 0;
#endif
		lutimes(name, times);
		chown(name, inode.di_uid, inode.di_gid);
	}

	umask(oldumask);
}

int
inodetofile(struct proginfo pi, struct dinode inode, const char *name)
{
	int fd;

	if ((fd = open(name, O_CREAT|O_WRONLY|O_TRUNC|O_EXLOCK,
	    inode.di_mode & 07777)) == -1)
		return -1;

	inodetofd(pi, inode, fd);

	close(fd);

	return 0;
}

void
inodetofd(struct proginfo pi, struct dinode inode, int fd)
{
	uint64_t i;
	ufs_daddr_t blktoget;

	for (i = 0; i <= (inode.di_size >> pi.fs->fs_bshift); i++) {
		if ((blktoget = getblknuminode(pi, inode, i)) != 0) {
			lseek(fd, lblktosize(pi.fs, i), SEEK_SET);
			if (i == lblkno(pi.fs, inode.di_size))
				writeout(pi, fd, getblkoff(pi, blktoget),
				    blkoff(pi.fs, inode.di_size));
			else
				writeout(pi, fd, getblkoff(pi, blktoget),
				    dblksize(pi.fs, &inode, i));
		}
	}
}

enum parts { P_DMIN, P_IMIN, P_SBLOCK, P_TOD };

int
isblkininode(struct proginfo pi, ufs_daddr_t blk, struct dinode inode)
{
	int found;
	ufs_daddr_t *blks;
	uint64_t i;

	if (blk == 0)
		return 0;

	found = 0;
	if ((inode.di_mode & IFMT) == IFREG) {
		/* check all first layer indirect blocks */
		for (i = 0; !found && i < 3; i++)
			if (blk == inode.di_ib[i])
				found = 1;

		/* check all second layer indirect blocks */
		blks = areadin(pi, inode.di_ib[1], pi.fs->fs_nindir
		    * sizeof *blks);
		for (i = 0; !found && i < pi.fs->fs_nindir; i++)
			if (blk == blks[i])
				found = 1;
		free(blks);

		/* XXX - no support for third layer indirect blocks */
		for (i = 0; !found && i <= (inode.di_size >> pi.fs->fs_bshift);
		    i++)
			if (blk == getblknuminode(pi, inode, i))
				found = 1;
	}

	return found;
}

void
findblk(struct proginfo pi, ufs_daddr_t blk, ino_t *num)
{
	int i;
	int diff;
	int tdiff;
	enum parts type;
	struct dinode inode;

	/* convert to file system block */
	blk = dbtofsb(pi.fs, blk);

	for (i = pi.fs->fs_ncg - 1; i >= 0 && cgbase(pi.fs, i) > blk; i--);

	diff = blk - cgdmin(pi.fs, i);
	type = P_DMIN;
	if (diff > (tdiff = blk - cgimin(pi.fs, i))) {
		diff = tdiff;
		type = P_IMIN;
	}
	if (diff > (tdiff = blk - cgtod(pi.fs, i))) {
		diff = tdiff;
		type = P_TOD;
	}
	if (diff > (tdiff = blk - cgsblock(pi.fs, i))) {
		diff = tdiff;
		type = P_SBLOCK;
	}

	switch (type) {
	case P_DMIN:
		printf("%d into data\n", blk);
		/* scan inodes to find block */
		printf("last: %d\n", pi.fs->fs_ncg * pi.fs->fs_ipg);
		for (i = 2; i < pi.fs->fs_ncg * pi.fs->fs_ipg; i++) {
			if (readin(pi, &inode, inumtooff(pi, i), sizeof inode)
			    != sizeof inode) {
				fprintf(stderr,
				    "short read on inode %d, skipping\n", i);
				continue;
			}
			if (isblkininode(pi, blk, inode)) {
				*num = i;
				printf("found in inum: %d\n", i);
				break;
			}
		}
		printf("end: %d\n", i);
		break;
	case P_IMIN:
		printf("%d into inode\n", blk);
		break;
	case P_SBLOCK:
		printf("%d into super\n", blk);
		break;
	case P_TOD:
		printf("%d into cylindar\n", blk);
		break;
	}
}

struct fs *
readinsb(struct proginfo pi, off_t blk)
{
	if (pi.map != NULL)
		return pi.map + blk * DEV_BSIZE;
	else
		return areadin(pi, blk * DEV_BSIZE, SBSIZE);
}

void
writeout(struct proginfo pi, int fd, off_t pos, size_t len)
{
	static ssize_t (*preadfun)(int fd, void *data, size_t len, off_t off)
	    = pread;
	static int buflen = 0;
	static void *buf = NULL;
	int i, j, k, l;

	if (pi.map != NULL) {
		i = 0;
		while (i < len) {
			if ((j = write(fd, pi.map + pos + i, len - i)) == -1) {
				perror("write");
				exit(1);
			}
			i += j;
		}
	} else {
		if (buflen < len) {
			if ((buf = realloc(buf, len)) == NULL) {
				perror("malloc");
				exit(1);
			}
			buflen = len;
		}

		i = 0;
		while (i < len) {
tryagain2:
			/*
			 * XXX - may return 0, but should never unless fs is
			 * corrupt.
			 */
			if ((j = preadfun(pi.fd, buf, len - i, pos + i))
			    == -1) {
				if (preadfun != bpread) {
					preadfun = bpread;
					goto tryagain2;
				}
				perror("pread");
				exit(1);
			}
			k = 0;
			while (k < j) {
				if ((l = write(fd, buf + k, j - k)) == -1) {
					perror("write");
					exit(1);
				}
				k += l;
			}
			i += j;
		}
	}
}

int
readin(struct proginfo pi, void *data, off_t pos, size_t len)
{
	static ssize_t (*preadfun)(int fd, void *data, size_t len, off_t off)
	    = pread;
	int i, j;

	if (pi.map != NULL) {
		memcpy(data, pi.map + pos, len);
		return len;
	} else {
		/* shift to the proper offset */
		pos += pi.off;
		i = 0;
		while (i < len) {
tryagain1:
			if ((j = preadfun(pi.fd, data + i, len - i, pos + i))
			    == -1) {
				if (preadfun != bpread) {
					preadfun = bpread;
					goto tryagain1;
				}
				perror("pread");
				exit(1);
			}
			if (j == 0)
				break;
			i += j;
		}
	}
	return i;
}

void *
areadin(struct proginfo pi, off_t pos, size_t len)
{
	void *ret;

	if ((ret = malloc(len)) == NULL) {
		perror("malloc");
		exit(1);
	}

	if (readin(pi, ret, pos, len) != len)
		fprintf(stderr,
		    "WARNING: short read on areadin, data may not be valid.\n");

	return ret;
}

ufs_daddr_t
getblknuminode(struct proginfo pi, struct dinode inode, uint64_t blk)
{
	ufs_daddr_t blktoget;

	if (blk < NDADDR)
		blktoget = inode.di_db[blk];
	else {
		/* should convert to a loop */
		if (blk >= NDADDR && blk < NINDIR(pi.fs) + NDADDR)
			readin(pi, &blktoget,
			    (off_t)(((ufs_daddr_t *)getblkoff(pi,
			    inode.di_ib[0])) + (blk - NDADDR)),
			    sizeof blktoget);
		else if (blk >= NINDIR(pi.fs) + NDADDR && blk
		    < pow(NINDIR(pi.fs), 2) + NDADDR) {
			readin(pi, &blktoget,
			    (off_t)(((ufs_daddr_t *)getblkoff(pi,
			    inode.di_ib[1])) + ((blk - NDADDR) /
			    pi.fs->fs_nindir)),
			    sizeof blktoget);
			if (blktoget != 0)
				readin(pi, &blktoget,
				    (off_t)(((ufs_daddr_t *)getblkoff(pi,
				    blktoget)) + ((blk - NDADDR)
				    % pi.fs->fs_nindir)),
				    sizeof blktoget);
		} else if (blk >= pow(pi.fs->fs_nindir, 2) + NDADDR && blk
		    < pow(pi.fs->fs_nindir, 3) + NDADDR) {
			fprintf(stderr,
			    "can't handle triple indirect blocks.\n");
			blktoget = 0;
		}
	}

	if (blktoget == 0)
		return 0;
	/* else */
	return blktoget;
}


void
printsuper(struct fs *fs)
{
	time_t gm;
	struct tm *tm;
	char *t;

	if ((t = validatesb(fs)) != NULL)
		printf("field %s does not match parent value.\n", t);

	printf("sblkno: %9d    cblkno: %7d    iblkno: %7d    dblkno: %7d\n",
	    fs->fs_sblkno, fs->fs_cblkno, fs->fs_iblkno, fs->fs_dblkno);
	printf("cgoffset: %7d    size: %9d    dsize: %8d    ncg: %10d\n",
	    fs->fs_cgoffset, fs->fs_size, fs->fs_dsize, fs->fs_ncg);
	printf("bsize: %10d    fsize: %8d    frag: %9d    minfree: %6d\n",
	    fs->fs_bsize, fs->fs_fsize, fs->fs_frag, fs->fs_minfree);
	printf("maxcontig: %6d    maxbpg: %7d    fsbtodb: %6d    nindir: %7d\n",
	    fs->fs_maxcontig, fs->fs_maxbpg, fs->fs_fsbtodb, fs->fs_nindir);
	printf("csaddr: %9d    cssize: %7d    cgsize: %7d    ncyl: %9d\n",
	    fs->fs_csaddr, fs->fs_cssize, fs->fs_cgsize, fs->fs_ncyl);
	printf("cpg: %12d    ipg: %10d    fpg: %10d    flags: %8x\n",
	    fs->fs_cpg, fs->fs_ipg, fs->fs_fpg, fs->fs_flags);
#ifdef FS_INODEFMT
	printf("inodefmt: %7d\n", fs->fs_inodefmt);
#endif
	printf("\nfsmnt: %s\n", fs->fs_fsmnt);
	gm = fs->fs_time;
	tm = gmtime(&gm);
	t = asctime(tm);
	t[strlen(t) - 1] = '\0';
	printf(  "time:  %s UTC\n", t);
}

char *
validatesb(struct fs *fs)
{
	if (fs->fs_bmask != ~(fs->fs_bsize - 1))
		return "fs_bmask";
	if (fs->fs_fmask != ~(fs->fs_fsize - 1))
		return "fs_fmask";
	if ((1 << fs->fs_bshift) != fs->fs_bsize)
		return "fs_bshift";
	if ((1 << fs->fs_fshift) != fs->fs_fsize)
		return "fs_fshift";
	if (fs->fs_fragshift != fs->fs_bshift - fs->fs_fshift)
		return "fs_fragshift";
	if ((DEV_BSIZE << fs->fs_fsbtodb) != fs->fs_fsize)
		return "fs_fsbtodb";
	/* need to check sbsize, csmask, csshift */
	if (fs->fs_nindir != fs->fs_bsize / sizeof(ufs_daddr_t))
		return "fs_nindir";
	if (fs->fs_inopb != fs->fs_bsize / sizeof(struct dinode))
		return "fs_inopb";
	if (fs->fs_nspf != fs->fs_fsize / DEV_BSIZE)
		return "fs_nspf";
	/* need to check csaddr, cssize, cgsize */
	/* need to check cpg, ipg, fpg */
#ifdef REAL_QUAD_T
	if (fs->fs_qbmask != fs->fs_bsize - 1)
		return "fs_qbmask";
	if (fs->fs_qfmask != fs->fs_fsize - 1)
		return "fs_qfmask";
#else
	if (*(int64_t *)&fs->fs_qbmask != fs->fs_bsize - 1)
		return "fs_qbmask";
	if (*(int64_t *)&fs->fs_qfmask != fs->fs_fsize - 1)
		return "fs_qfmask";
#endif

	return NULL;
}

ssize_t
bpread(int fd, void *data, size_t len, off_t off)
{
	char buf[DEV_BSIZE];
	off_t rndo;
	size_t rndl;
	size_t t;
	size_t pos;
	size_t olen;

	pos = 0;
	olen = len;
	rndo = off & -DEV_BSIZE;

	/* handle odd size first block, or front of block */
	if (off != rndo) {
		if ((pread(fd, buf, DEV_BSIZE, rndo)) == -1)
			return -1;
		if (len < DEV_BSIZE - (off - rndo))
			t = len;
		else
			t = DEV_BSIZE - (off - rndo);
		memcpy(data, buf + off - rndo, t);
		pos = t;
		len -= t;
		/* move to the next block */
		rndo += DEV_BSIZE;
	}

	/* handle middle block */
	rndl = len & -DEV_BSIZE;
	if (rndl != 0 && pread(fd, data + pos, rndl, rndo) == -1)
		return -1;
	pos += rndl;

	/* handle end block */
	if (rndl != len) {
		if (pread(fd, buf, DEV_BSIZE, rndo + rndl) == -1)
			return -1;
		memcpy(data + pos, buf, len - rndl);
	}

	return olen;
}
